"""
Core calculation functions and plotting routines.
"""

import os
_ROOT = os.path.abspath(os.path.dirname(__file__))
_excitation = os.path.join(_ROOT, 'excitation')
_interior = os.path.join(_ROOT, 'interior')
_induced = os.path.join(_ROOT, 'induced')
